#include<stdio.h>
#include<linux/in.h>
#include<netdb.h>
#include<sys/types.h>
#include<sys/time.h>
#include<sys/socket.h>
#include<stdlib.h>
#include<time.h>
#include<unistd.h>
#include<string.h>
#include"nfsd.h"
#include"lease.h"

void send_invalidation_message(long client,int file);
void wait_for_replies(struct lease_table_entry *file);
unsigned int max_timeout(struct lease_table_entry *file);
void clear_lease(struct lease_table_entry *file, unsigned long client);

struct lease_table_entry *lease_table_heads[HASH_SIZE];
struct soquete client_sockets[MAX_CLIENTS];
struct message msg;

int reply_sock = 0;


void initialize_lease_info()
{
 int i;
 struct sockaddr_in name;

 for(i=0; i <= MAX_SHARING; i++)
     lease_table_heads[i] = NULL;

 client_sockets[0].client = 0; /* No open sockets at the beggining. */
 gethostname(msg.hostname, sizeof(msg.hostname));
 msg.hostnamesize = strlen(msg.hostname);


 /* Let's creat a socket to receive the replies from the clients */ 

 if((reply_sock = socket(AF_INET, SOCK_DGRAM,0))<0)
   {
     dprintf(0,"nfsd: Couldn't open socket\n");
     exit(1);
   } 
      
 name.sin_family      = AF_INET;
 name.sin_addr.s_addr = htonl(INADDR_ANY);
 name.sin_port        = htons(SODA_SERVER_PORT);
 
 if(bind(reply_sock, (struct sockaddr*)&name, sizeof name)<0)
   {
     dprintf(0,"nfsd: Couldn't bind socket to port %d\n",SODA_SERVER_PORT);
     exit(1);
   } 
 
}/* initialize_lease_info */


/* In the future we'll need a function to free the space allocated for old 
   expired leases.
*/


struct lease_table_entry *find_lease(int file)
/* 
   Looks for this file on the lease_table. If it is not there, put it there. 
*/
{
  struct lease_table_entry *aux, *lease = lease_table_heads[hash_function(file)];
  
  
  while(lease && lease->psi_node != file && lease->colision_next)
    lease = lease->colision_next;
  
  if(lease && lease->psi_node == file) return lease;

  /* The entry was not found. So, let's create one... */
  
  aux = (struct lease_table_entry *) malloc(sizeof(struct lease_table_entry));
  aux->psi_node = file;
  aux->lease[0].client = 0;
  aux->colision_next = NULL;
  aux->colision_prev = lease;
  
  if(lease) lease->colision_next = aux;
  else      lease_table_heads[hash_function(file)] = aux;

  return aux;

}/* find_lease */


void new_lease(long client,int file,int expiration)
/* (client is the address of the client, 
   file is the local pseudo-inode number, and
   lease is the lease expiration time.)
   A new lease is inserted on the lease table or an old lease is extended.
*/
{
  int i=0;
  struct lease_table_entry *entry = find_lease(file);

  while(entry->lease[i].client && entry->lease[i].client != client)
    i++; 

  if(!entry->lease[i].client) 
    { 
      if(i>MAX_SHARING) 
	{
	  dprintf(0,"It seems that you have more than %d clients sharing a file. \n" 
		  "YOU MUST RECOMPILE nfsd WITH BIGGER MAX_SHARING\n",MAX_SHARING); 
	  exit(1); 
	} 
      entry->lease[i].client = client; 
      entry->lease[i+1].client = 0; 
/*      dprintf(0,"Inseri novo cliente: %d\n",client); */
    } 
/*  else dprintf(0,"Aproveitei o lease antigo (%d) do cliente: %X\n", 
	       entry->lease[i].expiration, client);
*/
  entry->lease[i].expiration = expiration; 
}/* new_lease */



void invalidate_leases(int file)
/* 
   Looks for all the clients that have a valid lease for this file, send invalidation
   messages to them and wait for the replies or for the lease time-outs.
*/
{
  int i, message_sent = 0;
  
  struct lease_table_entry *entry = lease_table_heads[hash_function(file)];
  
  /* dprintf(0,"vou invalidar os leases do arquivo %d\n",file); */

  while(entry && entry->psi_node != file)
    entry = entry->colision_next;
  
  if(!entry) 
    {
     /* dprintf(0,"Nao achei nenhum lease para esse arquivo\n"); */
      return; /* There are no leases for this file. */
    }

  
  for(i=0; entry->lease[i].client; i++)
    {    
      if(entry->lease[i].expiration > time(NULL))
	{
	  message_sent =1;
	  send_invalidation_message(entry->lease[i].client,file);
	  /* dprintf(0,"Mandei mensagem para cliente %X\n",entry->lease[i].client);*/
	}
    }

  /* dprintf(0,"%d Tratei de %d leases\n",i); */

  if(message_sent) wait_for_replies(entry);

}/* invalidate_leases */


/**************************************************************************
   The following functions manage the communication sockets which are
  used to talk to the clients.
**************************************************************************/


void send_invalidation_message(long client,int file)
{
  int i,sock;
  struct sockaddr_in name;
  struct hostent *hp;
  
  /* Looks for a socket connected to this client. */
  for(i=0; client_sockets[i].client && client_sockets[i].client != client; i++);
  
  if(!client_sockets[i].client) /* We need to connect a socket to the client */
    {
      /* dprintf(0,"Vou abrir soquete para o cliente %X\n",client); */
      if(i>MAX_CLIENTS) 
	{
	  dprintf(0,"It seems that you have more than %d clients. \n"
		 "YOU MUST RECOMPILE nfsd WITH BIGGER MAX_CLIENTS\n",MAX_CLIENTS);
	  exit(1);
	}

      if((sock = socket(AF_INET, SOCK_DGRAM,0))<0)
	{
	  dprintf(0,"Couldn't open socket\n");
	  return;
	}
      if((hp = gethostbyaddr((char *) &client, sizeof client, AF_INET))==0)
	{
	  dprintf(0,"Client %X is unreachable.\n",client);
	  return;
	}
  
      bcopy((char *)hp->h_addr, (char *)&name.sin_addr, hp->h_length);
      name.sin_family = AF_INET;
      name.sin_port = htons(SODA_CLIENT_PORT);
      
      if(connect(sock,(struct sockaddr *)&name, sizeof name)<0)
	{
	  dprintf(0,"Couldn't connect to client %X\n",client);
	  return;
	}

      client_sockets[i].client   = client;
      client_sockets[i].sock     = sock;
      client_sockets[i+1].client = 0;
    }

/* 
   verificar se file e' igual ao i-node que o cliente guarda!!! Acho
   que e'; verificar atraves de testes.
*/

  msg.file = file;
  send(client_sockets[i].sock, &msg, ( 2*sizeof(int) + msg.hostnamesize + 1), 0);

}/* send_invalidation_message */



void wait_for_replies(struct lease_table_entry *file)
{
  static struct sockaddr_in from;
  static fd_set ready;

  int result, fromlen;
  struct timeval timeout;
  int psi_node;             /* pseudo-i-node of the file */

  /* Once we are single-threaded, we cannot treat other request while
     this write is not commited. We'll sleep until some message comes
     or a timeout happens. Good night...
  */

  while(1)
    {
      timeout.tv_usec = 0;
      if (timeout.tv_sec = max_timeout(file))
	{
	  FD_ZERO(&ready);
	  FD_SET(reply_sock, &ready);
	  
	  result = select(FD_SETSIZE, &ready, (fd_set *) 0, (fd_set *) 0, &timeout);
	}      

      /* When a timeout occurs, result = 0 */
      
      if(!result || !timeout.tv_sec) 
	{
	  /* dprintf(0,"timeout\n"); */
	  return; 
	}

      if(!FD_ISSET(reply_sock, &ready)) 
	{
	  dprintf(0,"nfsd:Internal error after select() call\n");
	  exit(1);
	}
      else /* There's a message on reply_sock. */
	{
	  fromlen = sizeof from; /* this line must be HERE! */
	  recvfrom(reply_sock, &psi_node, sizeof psi_node, 0,(struct sockaddr *) &from, 
		   &fromlen);
	  /* estou ignorando o possivel erro quando o valor -1 e' retornado. */

	  if (psi_node == file->psi_node)
	    {
	      /* dprintf(0,"Recebi resposta do cliente %X\n",from.sin_addr.s_addr); */
	      clear_lease(file,from.sin_addr.s_addr);
	    }
	  else
	   dprintf(1,"Unexpected message has arrived from client %X\n",
		   from.sin_addr.s_addr);
	}
    }/* while(1) */
}/* wait_for_replies */



unsigned int max_timeout(struct lease_table_entry *file)
{
 int i, max = 0;

 for(i=0; file->lease[i].client; i++)
   if(file->lease[i].expiration > max)
     max = file->lease[i].expiration;

 max = max - time(NULL);

 /* dprintf(0,"Vou esperar ate' %d segundos\n",(max<0)? 0 : max); */

 return (max<0)? 0 : max;
} 



void clear_lease(struct lease_table_entry *file, unsigned long client)
{
 int i = 0;

 while(file->lease[i].client && file->lease[i].client != client)
   i++; /* Looks for this lease. */

 if(file->lease[i].client)
   {
     /* We shift left one position all subsequent leases.*/
     /* We can afford such an ugly thing because we expect i to be small. */
     while(file->lease[i].client)
       {
	 file->lease[i].client     = file->lease[i+1].client;
	 file->lease[i].expiration = file->lease[i++].expiration;
       }
   }
}






