/* lease.c : contains revalidate() and sys_invalidate_lease() */

#include<linux/string.h>
#include<linux/kernel.h>
#include"cache.h"
#include"structures.h"	
#include"hash.h"
#include<asm/segment.h>


#define MAX_HOSTNAME 256

extern int sys_time(void *);


int revalidate_lease(unsigned long inode_number, char *hostname, 
		     unsigned long new_lease, char *data, off_t offset, 
		     int count)
/*
  It's called when an old lease is validated for another period.
  We copy the cached block to *data.
*/
{
  int size, last_byte;
  struct hash_table_entry *file;
  struct list_node *block;

  if (!(file  = find_hash_entry(inode_number, hostname)))
    {
      PRINTK("The file was flushed out from the cache while the RPC was made.\n");
      return 0; 
    }           

  /* if (file->lease_expiration < new_lease)  Just in case... */
    file->lease_expiration = new_lease;   
  
/* We'll do an awful thing now: we'll look for the cached block again. 
*/

  if(!(block = find_cache_block(file, offset)) ||
     block->offset + block->size <= offset)
    {
      PRINTK("What a hell, the data aren't cached anymore :-(\n");
      return 0;
    }
/*d
   We'll not check if the lease came expired while making tests about the SPRITE protocol.


  if (file->lease_expiration < sys_time(NULL))
    {
      PRINTK("What a hell, the lease came expired :-(\n");
      return 0;
    }
*/

  last_byte = block->offset + block->size;
  if (offset + count < last_byte) last_byte = offset + count;
  size = last_byte - offset;
  memcpy(data, block->data + (offset - block->offset), size);
    
  refresh_block_LRU_list(block);

 return 1;                               
}



asmlinkage int sys_invalidate_lease(unsigned int file, char * hostname, int size)
/*
   This is the system call called by the sodad daemon.
*/
{
  struct hash_table_entry *entry;
  char host[MAX_HOSTNAME];

  memcpy_fromfs(host, hostname, size+1); /* copy from segment fs */

  entry = find_hash_entry(file, host);

  if (entry) 
    {
      entry->lease_expiration = 0;
      PRINTK("invalidei lease\n");
    }
  else
    PRINTK("nao encontrei o lease\n");

  return 0;
}

